@@ -2,7 +2,7 @@
  * @file   core.c
  * @author Wei-Ning Huang (AZ) <aitjcize@gmail.com>
  *
- * Copyright (C) 2013 -  Wei-Ning Huang (AZ) <aitjcize@gmail.com>
+ * Copyright (C) 2013 - 2014  Wei-Ning Huang (AZ) <aitjcize@gmail.com>
  * All Rights reserved.
  *
  * This program is free software; you can redistribute it and/or modify
@@ -31,6 +31,16 @@ PyObject* ToxCoreError;
     return NULL;                                               \
   }
 
+#if PY_MAJOR_VERSION < 3
+  #define PYSTRING_FromString PyString_FromString
+  #define PYSTRING_FromStringAndSize PyString_FromStringAndSize
+  #define PYSTRING_Check PyString_Check
+#else
+  #define PYSTRING_FromString PyUnicode_FromString
+  #define PYSTRING_FromStringAndSize PyBytes_FromStringAndSize
+  #define PYSTRING_Check PyUnicode_Check
+#endif
+
 static void bytes_to_hex_string(uint8_t* digest, int length,
     uint8_t* hex_digest)
 {
@@ -72,28 +82,19 @@ static void hex_string_to_bytes(uint8_t* hexstr, int length, uint8_t* bytes)
   }
 }
 
-/* Helper functions for Python 2 and 3 compability */
-
-// PyString_ functions were deprecated in Python 3 in favour
-// of PyUnicode_for string data and PyBytes_ for binary data:
-PyObject* PyUnicodeString_FromString(const char *str) {
-  PyObject* res;
+void PyStringUnicode_AsStringAnsSize(PyObject* object, char** str,
+    Py_ssize_t* len)
+{
 #if PY_MAJOR_VERSION < 3
-  res = PyString_FromString(str);
+    PyString_AsStringAndSize(object, str, len);
 #else
-  res = PyUnicode_FromString(str);
+# if PY_MINOR_VERSION == 2
+    *str = PyUnicode_AS_DATA(object);
+    *len = PyUnicode_GET_DATA_SIZE(object);
+# else
+    *str = PyUnicode_AsUTF8AndSize(object, len);
+# endif
 #endif
-  return res;
-}
-
-PyObject* PyByteString_FromString(const char *str) {
-  PyObject* res;
-#if PY_MAJOR_VERSION < 3
-  res = PyString_FromString(str);
-#else
-  res = PyBytes_FromString(str);
-#endif
-  return res;  
 }
 
 /* core.Tox definition */
@@ -111,35 +112,35 @@ static void callback_friend_request(uint8_t* public_key, uint8_t* data,
   bytes_to_hex_string(public_key, TOX_CLIENT_ID_SIZE, buf);
 
   PyObject_CallMethod((PyObject*)self, "on_friend_request", "ss#", buf, data,
-      length - 1);
+      length > 0? length - 1: 0);
 }
 
 static void callback_friend_message(Tox *tox, int friendnumber,
     uint8_t* message, uint16_t length, void* self)
 {
   PyObject_CallMethod((PyObject*)self, "on_friend_message", "is#", friendnumber,
-      message, length - 1);
+      message, length > 0? length - 1: 0);
 }
 
 static void callback_action(Tox *tox, int friendnumber, uint8_t* action,
     uint16_t length, void* self)
 {
-  PyObject_CallMethod((PyObject*)self, "on_action", "is#", friendnumber, action,
-      length - 1);
+  PyObject_CallMethod((PyObject*)self, "on_friend_action", "is#", friendnumber,
+      action, length > 0? length - 1: 0);
 }
 
 static void callback_name_change(Tox *tox, int friendnumber, uint8_t* newname,
     uint16_t length, void* self)
 {
   PyObject_CallMethod((PyObject*)self, "on_name_change", "is#", friendnumber,
-      newname, length - 1);
+      newname, length > 0? length - 1: 0);
 }
 
 static void callback_status_message(Tox *tox, int friendnumber,
     uint8_t *newstatus, uint16_t length, void* self)
 {
   PyObject_CallMethod((PyObject*)self, "on_status_message", "is#", friendnumber,
-      newstatus, length - 1);
+      newstatus, length > 0? length - 1: 0);
 }
 
 static void callback_user_status(Tox *tox, int friendnumber,
@@ -178,7 +179,14 @@ static void callback_group_message(Tox *tox, int groupid,
     int friendgroupid, uint8_t* message, uint16_t length, void *self)
 {
   PyObject_CallMethod((PyObject*)self, "on_group_message", "iis#", groupid,
-      friendgroupid, message, length - 1);
+      friendgroupid, message, length > 0? length - 1: 0);
+}
+
+static void callback_group_action(Tox *tox, int groupid,
+    int friendgroupid, uint8_t* action, uint16_t length, void *self)
+{
+  PyObject_CallMethod((PyObject*)self, "on_group_action", "iis#", groupid,
+      friendgroupid, action, length > 0? length - 1: 0);
 }
 
 static void callback_group_namelist_change(Tox *tox, int groupid,
@@ -193,7 +201,8 @@ static void callback_file_send_request(Tox *m, int friendnumber,
     uint16_t filename_length, void* self)
 {
   PyObject_CallMethod((PyObject*)self, "on_file_send_request", "iiKs#",
-      friendnumber, filenumber, filesize, filename, filename_length);
+      friendnumber, filenumber, filesize, filename,
+      filename_length > 0? filename_length - 1: 0);
 }
 
 static void callback_file_control(Tox *m, int friendnumber,
@@ -238,10 +247,20 @@ static int init_helper(ToxCore* self, PyObject* args)
   }
 
   Tox* tox = tox_new(ipv6enabled);
+  if (tox == NULL) {
+      fprintf(stderr, "Warning: failed to initialize toxcore with ipv6, "
+          "trying ipv4.\n");
+
+      tox = tox_new(0);
+      if (tox == NULL) {
+        PyErr_SetString(ToxCoreError, "failed to initialize toxcore");
+        return -1;
+      }
+  }
 
   tox_callback_friend_request(tox, callback_friend_request, self);
   tox_callback_friend_message(tox, callback_friend_message, self);
-  tox_callback_action(tox, callback_action, self);
+  tox_callback_friend_action(tox, callback_action, self);
   tox_callback_name_change(tox, callback_name_change, self);
   tox_callback_status_message(tox, callback_status_message, self);
   tox_callback_user_status(tox, callback_user_status, self);
@@ -249,6 +268,7 @@ static int init_helper(ToxCore* self, PyObject* args)
   tox_callback_connection_status(tox, callback_connection_status, self);
   tox_callback_group_invite(tox, callback_group_invite, self);
   tox_callback_group_message(tox, callback_group_message, self);
+  tox_callback_group_action(tox, callback_group_action, self);
   tox_callback_group_namelist_change(tox, callback_group_namelist_change, self);
   tox_callback_file_send_request(tox, callback_file_send_request, self);
   tox_callback_file_control(tox, callback_file_control, self);
@@ -310,8 +330,7 @@ ToxCore_getaddress(ToxCore* self, PyObject* args)
   tox_get_address(self->tox, address);
   bytes_to_hex_string(address, TOX_FRIEND_ADDRESS_SIZE, address_hex);
 
-  PyObject* res = PyUnicodeString_FromString((const char*)address_hex);
-  return res;
+  return PYSTRING_FromString((const char*)address_hex);
 }
 
 static PyObject*
@@ -441,7 +460,7 @@ ToxCore_get_client_id(ToxCore* self, PyObject* args)
   tox_get_client_id(self->tox, friendid, pk);
   bytes_to_hex_string(pk, TOX_CLIENT_ID_SIZE, hex);
 
-  return PyUnicodeString_FromString((const char*)hex);
+  return PYSTRING_FromString((const char*)hex);
 }
 
 static PyObject*
@@ -500,7 +519,7 @@ ToxCore_friend_exists(ToxCore* self, PyObject* args)
 }
 
 static PyObject*
-ToxCore_sendmessage(ToxCore* self, PyObject* args)
+ToxCore_send_message(ToxCore* self, PyObject* args)
 {
   CHECK_TOX(self);
 
@@ -522,7 +541,7 @@ ToxCore_sendmessage(ToxCore* self, PyObject* args)
 }
 
 static PyObject*
-ToxCore_sendmessage_withid(ToxCore* self, PyObject* args)
+ToxCore_send_message_withid(ToxCore* self, PyObject* args)
 {
   CHECK_TOX(self);
 
@@ -624,7 +643,7 @@ ToxCore_get_self_name(ToxCore* self, PyObject* args)
     return NULL;
   }
 
-  return PyUnicodeString_FromString((const char*)buf);
+  return PYSTRING_FromString((const char*)buf);
 }
 
 static PyObject*
@@ -645,7 +664,7 @@ ToxCore_get_name(ToxCore* self, PyObject* args)
     return NULL;
   }
 
-  return PyUnicodeString_FromString((const char*)buf);
+  return PYSTRING_FromString((const char*)buf);
 }
 
 static PyObject*
@@ -725,7 +744,7 @@ ToxCore_get_status_message(ToxCore* self, PyObject* args)
 
   buf[TOX_MAX_STATUSMESSAGE_LENGTH -1] = 0;
 
-  return PyUnicodeString_FromString((const char*)buf);
+  return PYSTRING_FromString((const char*)buf);
 }
 
 static PyObject*
@@ -746,7 +765,7 @@ ToxCore_get_self_status_message(ToxCore* self, PyObject* args)
 
   buf[TOX_MAX_STATUSMESSAGE_LENGTH -1] = 0;
 
-  return PyUnicodeString_FromString((const char*)buf);
+  return PYSTRING_FromString((const char*)buf);
 }
 
 static PyObject*
@@ -799,6 +818,15 @@ ToxCore_count_friendlist(ToxCore* self, PyObject* args)
   return PyLong_FromUnsignedLong(count);
 }
 
+static PyObject*
+ToxCore_get_num_online_friends(ToxCore* self, PyObject* args)
+{
+  CHECK_TOX(self);
+
+  uint32_t count = tox_get_num_online_friends(self->tox);
+  return PyLong_FromUnsignedLong(count);
+}
+
 static PyObject*
 ToxCore_get_friendlist(ToxCore* self, PyObject* args)
 {
@@ -873,7 +901,7 @@ ToxCore_group_peername(ToxCore* self, PyObject* args)
     PyErr_SetString(ToxCoreError, "failed to get group peername");
   }
 
-  return PyUnicodeString_FromString((const char*)buf);
+  return PYSTRING_FromString((const char*)buf);
 }
 
 static PyObject*
@@ -938,6 +966,26 @@ ToxCore_group_message_send(ToxCore* self, PyObject* args)
   Py_RETURN_NONE;
 }
 
+static PyObject*
+ToxCore_group_action_send(ToxCore* self, PyObject* args)
+{
+  CHECK_TOX(self);
+
+  int groupid = 0;
+  uint8_t* action = NULL;
+  uint32_t length = 0;
+
+  if (!PyArg_ParseTuple(args, "is#", &groupid, &action, &length)) {
+    return NULL;
+  }
+
+  if (tox_group_action_send(self->tox, groupid, action, length + 1) == -1) {
+    PyErr_SetString(ToxCoreError, "failed to send group action");
+  }
+
+  Py_RETURN_NONE;
+}
+
 static PyObject*
 ToxCore_group_number_peers(ToxCore* self, PyObject* args)
 {
@@ -981,7 +1029,7 @@ ToxCore_group_get_names(ToxCore* self, PyObject* args)
 
   int i = 0;
   for (i = 0; i < n2; ++i) {
-    PyList_Append(list, PyUnicodeString_FromString((const char*)names[i]));
+    PyList_Append(list, PYSTRING_FromString((const char*)names[i]));
   }
 
   free(names);
@@ -1028,7 +1076,7 @@ ToxCore_new_filesender(ToxCore* self, PyObject* args)
   CHECK_TOX(self);
 
   int friendnumber = 0;
-  int filesize = 0;
+  uint64_t filesize = 0;
   uint8_t* filename = 0;
   int filename_length = 0;
 
@@ -1038,7 +1086,7 @@ ToxCore_new_filesender(ToxCore* self, PyObject* args)
   }
 
   int ret = tox_new_file_sender(self->tox, friendnumber, filesize, filename,
-      filename_length);
+      filename_length + 1);
 
   if (ret == -1) {
     PyErr_SetString(ToxCoreError, "tox_new_file_sender() failed");
@@ -1231,23 +1279,44 @@ ToxCore_save(ToxCore* self, PyObject* args)
 {
   CHECK_TOX(self);
 
-  PyObject* res = NULL;
-  int size = tox_size(self->tox);
+  PyObject* keyobj = Py_None;
+  uint8_t* key = NULL;
+  Py_ssize_t key_len = 0;
+
+  if (!PyArg_ParseTuple(args, "|O", &keyobj)) {
+    return NULL;
+  }
+
+  if (PYSTRING_Check(keyobj)) {
+    PyStringUnicode_AsStringAnsSize(keyobj, (char**)&key, &key_len);
+  } else if (keyobj != Py_None) {
+    PyErr_SetString(ToxCoreError, "invalid passphrase type");
+    return NULL;
+  }
+
+  int size = 0;
+  if (key) {
+    size = tox_size_encrypted(self->tox);
+  } else {
+    size = tox_size(self->tox);
+  }
   uint8_t* buf = (uint8_t*)malloc(size);
 
   if (buf == NULL) {
     return PyErr_NoMemory();
   }
 
-  tox_save(self->tox, buf);
+  if (key) {
+    if (tox_save_encrypted(self->tox, buf, key, key_len) == -1) {
+      PyErr_SetString(ToxCoreError, "tox_save_encrypted(): failed");
+      free(buf);
+      return NULL;
+    }
+  } else {
+    tox_save(self->tox, buf);
+  }
 
-  // The PyString_* functions were split in Python 3.* into
-  // PyUnicode_ for string data and PyBytes_ for binary data
-#if PY_MAJOR_VERSION < 3
-  res = PyString_FromStringAndSize((const char*)buf, size);
-#else
-  res = PyBytes_FromStringAndSize((const char*)buf, size);
-#endif
+  PyObject* res = PYSTRING_FromStringAndSize((const char*)buf, size);
   free(buf);
 
   return res;
@@ -1258,24 +1327,49 @@ ToxCore_save_to_file(ToxCore* self, PyObject* args)
 {
   CHECK_TOX(self);
 
-  char* filename;
-  if (!PyArg_ParseTuple(args, "s", &filename)) {
+  PyObject* keyobj = Py_None;
+  char* filename = NULL;
+  uint8_t* key = NULL;
+  Py_ssize_t key_len = 0;
+
+  if (!PyArg_ParseTuple(args, "s|O", &filename, &keyobj)) {
     return NULL;
   }
 
-  int size = tox_size(self->tox);
+  if (PYSTRING_Check(keyobj)) {
+    PyStringUnicode_AsStringAnsSize(keyobj, (char**)&key, &key_len);
+  } else if (keyobj != Py_None) {
+    PyErr_SetString(ToxCoreError, "invalid passphrase type");
+    return NULL;
+  }
+
+  int size = 0;
+  if (key) {
+    size = tox_size_encrypted(self->tox);
+  } else {
+    size = tox_size(self->tox);
+  }
+
   uint8_t* buf = (uint8_t*)malloc(size);
 
   if (buf == NULL) {
     return PyErr_NoMemory();
   }
 
-  tox_save(self->tox, buf);
+  if (key) {
+    if (tox_save_encrypted(self->tox, buf, key, key_len) == -1) {
+      PyErr_SetString(ToxCoreError, "tox_save_encrypted(): failed");
+      free(buf);
+      return NULL;
+    }
+  } else {
+    tox_save(self->tox, buf);
+  }
 
   FILE* fp = fopen(filename, "w");
 
   if (fp == NULL) {
-    PyErr_SetString(ToxCoreError, "tox_save(): can't open file for saving");
+    PyErr_SetString(ToxCoreError, "fopen(): can't open file for saving");
     return NULL;
   }
 
@@ -1294,17 +1388,35 @@ ToxCore_load(ToxCore* self, PyObject* args)
 
   uint8_t* data = NULL;
   int length = 0;
+  PyObject* keyobj = Py_None;
+  uint8_t* key = NULL;
+  Py_ssize_t key_len = 0;
 
-  if (!PyArg_ParseTuple(args, "s#", &data, &length)) {
+  if (!PyArg_ParseTuple(args, "s#|O", &data, &length, &keyobj)) {
     PyErr_SetString(ToxCoreError, "no data specified");
     return NULL;
   }
 
-  if (tox_load(self->tox, data, length) == -1) {
-    Py_RETURN_FALSE;
-  } else {
-    Py_RETURN_TRUE;
+  if (PYSTRING_Check(keyobj)) {
+    PyStringUnicode_AsStringAnsSize(keyobj, (char**)&key, &key_len);
+  } else if (keyobj != Py_None) {
+    PyErr_SetString(ToxCoreError, "invalid passphrase type");
+    return NULL;
   }
+
+  if (key) {
+    if (tox_load_encrypted(self->tox, data, length, key, key_len) == -1) {
+      PyErr_SetString(ToxCoreError, "tox_load_encrypted(): load failed");
+      return NULL;
+    }
+  } else {
+    if (tox_load(self->tox, data, length) == -1) {
+      PyErr_SetString(ToxCoreError, "tox_load(): load failed");
+      return NULL;
+    }
+  }
+
+  Py_RETURN_NONE;
 }
 
 static PyObject*
@@ -1313,8 +1425,18 @@ ToxCore_load_from_file(ToxCore* self, PyObject* args)
   CHECK_TOX(self);
 
   char* filename = NULL;
+  PyObject* keyobj = Py_None;
+  uint8_t* key = NULL;
+  Py_ssize_t key_len = 0;
 
-  if (!PyArg_ParseTuple(args, "s", &filename)) {
+  if (!PyArg_ParseTuple(args, "s|O", &filename, &keyobj)) {
+    return NULL;
+  }
+
+  if (PYSTRING_Check(keyobj)) {
+    PyStringUnicode_AsStringAnsSize(keyobj, (char**)&key, &key_len);
+  } else if (keyobj != Py_None) {
+    PyErr_SetString(ToxCoreError, "invalid passphrase type");
     return NULL;
   }
 
@@ -1335,14 +1457,26 @@ ToxCore_load_from_file(ToxCore* self, PyObject* args)
   if (fread(data, length, 1, fp) != 1) {
     PyErr_SetString(ToxCoreError,
         "tox_load(): corrupted data file");
+    free(data);
     return NULL;
   }
 
-  if (tox_load(self->tox, data, length) == -1) {
-    Py_RETURN_FALSE;
+  if (key) {
+    if (tox_load_encrypted(self->tox, data, length, key, key_len) == -1) {
+      PyErr_SetString(ToxCoreError, "tox_load_encrypted(): load failed");
+      free(data);
+      return NULL;
+    }
   } else {
-    Py_RETURN_TRUE;
+    if (tox_load(self->tox, data, length) == -1) {
+      PyErr_SetString(ToxCoreError, "tox_load(): load failed");
+      free(data);
+      return NULL;
+    }
   }
+
+  free(data);
+  Py_RETURN_NONE;
 }
 
 PyMethodDef Tox_methods[] = {
@@ -1359,8 +1493,8 @@ PyMethodDef Tox_methods[] = {
     "nothing."
   },
   {
-    "on_action", (PyCFunction)ToxCore_callback_stub, METH_VARARGS,
-    "on_action(friend_number, action)\n"
+    "on_friend_action", (PyCFunction)ToxCore_callback_stub, METH_VARARGS,
+    "on_friend_action(friend_number, action)\n"
     "Callback for receiving friend actions, default implementation does "
     "nothing."
   },
@@ -1410,6 +1544,12 @@ PyMethodDef Tox_methods[] = {
     "Callback for receiving group messages, default implementation does "
     "nothing."
   },
+  {
+    "on_group_action", (PyCFunction)ToxCore_callback_stub, METH_VARARGS,
+    "on_group_action(group_number, friend_group_number, action)\n"
+    "Callback for receiving group actions, default implementation does "
+    "nothing."
+  },
   {
     "on_group_namelist_change", (PyCFunction)ToxCore_callback_stub,
     METH_VARARGS,
@@ -1491,12 +1631,12 @@ PyMethodDef Tox_methods[] = {
     "Checks if there exists a friend with given friendnumber."
   },
   {
-    "send_message", (PyCFunction)ToxCore_sendmessage, METH_VARARGS,
+    "send_message", (PyCFunction)ToxCore_send_message, METH_VARARGS,
     "send_message(friend_number, message)\n"
     "Send a text chat message to an online friend."
   },
   {
-    "send_message_withid", (PyCFunction)ToxCore_sendmessage_withid,
+    "send_message_withid", (PyCFunction)ToxCore_send_message_withid,
     METH_VARARGS,
     "send_message_withid(friend_number, id, message)\n"
     "Send a text chat message to an online friend with id."
@@ -1594,6 +1734,12 @@ PyMethodDef Tox_methods[] = {
     "count_friendlist()\n"
     "Return the number of friends."
   },
+  {
+    "get_num_online_friends", (PyCFunction)ToxCore_get_num_online_friends,
+    METH_NOARGS,
+    "get_num_online_friends()\n"
+    "Return the number of online friends."
+  },
   {
     "get_friendlist", (PyCFunction)ToxCore_get_friendlist,
     METH_NOARGS,
@@ -1630,6 +1776,11 @@ PyMethodDef Tox_methods[] = {
     "group_message_send(group_number, message)\n"
     "send a group message."
   },
+  {
+    "group_action_send", (PyCFunction)ToxCore_group_action_send, METH_VARARGS,
+    "group_action_send(group_number, action)\n"
+    "send a group action."
+  },
   {
     "group_number_peers", (PyCFunction)ToxCore_group_number_peers, METH_VARARGS,
     "group_number_peers(group_number)\n"
@@ -1721,23 +1872,23 @@ PyMethodDef Tox_methods[] = {
     "return size of messenger data (for saving)."
   },
   {
-    "save", (PyCFunction)ToxCore_save, METH_NOARGS,
-    "save()\n"
+    "save", (PyCFunction)ToxCore_save, METH_VARARGS,
+    "save([passphrase=None])\n"
     "Return messenger blob in str."
   },
   {
     "save_to_file", (PyCFunction)ToxCore_save_to_file, METH_VARARGS,
-    "save_to_file(filename)\n"
+    "save_to_file(filename[, passphrase=None])\n"
     "Save the messenger to a file."
   },
   {
     "load", (PyCFunction)ToxCore_load, METH_VARARGS,
-    "load(blob)\n"
+    "load(blob[, passphrase=None])\n"
     "Load the messenger from *blob*."
   },
   {
     "load_from_file", (PyCFunction)ToxCore_load_from_file, METH_VARARGS,
-    "load_from_file(filename)\n"
+    "load_from_file(filename[, passphrase=None])\n"
     "Load the messenger from file."
   },
   {NULL}
